from distutils.core import setup

setup(
    name='pssm',
    packages=['pssm'],
    version='0.0.14',
    description='Python library for state-space models',
    author='Rui Vieira',
    author_email='ruidevieira@googlemail.com',
    url='https://github.com/ruivieira/python-ssm',
    download_url='https://github.com/'
                 'ruivieira/python-ssm/archive/0.0.14.tar.gz',
    keywords=['ssm', 'statistics', 'scientific', 'state-space', 'models'],
    classifiers=[],
)
